#include "mns.h"

#define OSCFGFILE "fingerprints.conf"

#define SERVICES 24
#define ACK_RAND 200
#define PACKETS 8
#define PORT_CLOSED 1

#define TH_XXX 0x40
#define TH_YYY 0x80

#define iph ((struct ip *)pkt)
#define icmph ((struct icmphdr *)&pkt[iph->ip_hl << 2])
#define tcph ((struct tcphdr *)&pkt[iph->ip_hl << 2])
#define MIN_PKT 200

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

struct servicetype {
   char *name;
   int port;
};

typedef struct {
   u_long seq;
   u_long ack;
} hdrinfo;

/* function prototypes */
void rawsockets(int *, int *);
char *checkosmain(int, int);
void sendtcp(int, hdrinfo *, u_short);
int synscan(int, int, int *, hdrinfo);
int finscan(int, int, int *, hdrinfo);
u_long gettcp(int, struct tcphdr *);
struct in_addr targetup(int);

/* global variables */
struct in_addr local_ip;
struct in_addr target_ip;
time_t timedelay;
u_short src_port;
u_short dst_port;

static struct servicetype service[SERVICES] = {
   { "echo", 7 },
   { "systat", 11 },
   { "daytime", 13 },
   { "netstat", 15 },
   { "chargen", 19 },
   { "ftp", 21 },
   { "ssh", 22 },
   { "telnet", 23 },
   { "smtp", 25 },
   { "time", 37 },
   { "domain", 53 },
   { "finger", 79 },
   { "www", 80 },
   { "kerberos", 88 },
   { "pop3", 110 },
   { "sunrcp", 111 },
   { "auth", 113 },
   { "nntp", 119 },
   { "netbios-ns", 137 },
   { "netbios-ssn", 139 },
   { "imap", 143 },
   { "login", 513 },
   { "shell", 514 },
   { "printer", 515 }
};

struct node *stealthscan (struct in_addr target,
                          struct node *walk,
                          int **plist, struct in_addr *hostaddress)
{
   int (*scanfunc)(int, int, int *, hdrinfo);
   hdrinfo seq_ack;
   struct tcphdr itcp;
   int i, j, k;
   int sd_send, sd_recv;
   int pschkd[20] = { 0 };
   time_t timeout;
   long ack_diff = 0;
   char buffer[80];

   srand (time (NULL));

   rawsockets (&sd_send, &sd_recv);

   target_ip = target;
   local_ip = targetup(sd_recv);
   
   if(!local_ip.s_addr) {
      stralloc(walk, "\n\t- no response from target machine");
      *plist = NULL;
      return walk;
   } else *hostaddress = local_ip;

   /* this is a TH_SYN | TH_FIN check that happens to be are really useful
    * check.  If an ACK of +2 is received then we know almost for sure
    * that we're looking at a Windows based system.
    * If we get no response, we either hit an open port on a *NIX based
    * machine or the machine is firewalled.
    * If we get ACK +2 or hit a firewalled machine, FIN scan won't work.
    * As far as I know, I'm the first one who has come up with this test.
    */
    
   seq_ack.ack = 0;
   src_port = htons((rand() % 10000) + 5000);
   seq_ack.seq = rand();
   
   sendtcp(sd_send, &seq_ack, TH_SYN | TH_FIN);

   for(i = j = 0; i < 2; i++) { /* we try twice just to make sure */
      timeout = time(NULL); /* we also calibrate the time delay */
      while(timeout + TIMEOUT > time(NULL))
         if(gettcp(sd_recv, &itcp))
            if(src_port == itcp.th_dport) { /* response to our request? */
               if(ntohl(itcp.th_ack)) /* did we get an ACK? */
                  ack_diff = ntohl(itcp.th_ack) - seq_ack.seq; /* ACK-SEQ */
               j = 1; /* yes, we got a valid packet */
               break;
            }
      if(j) break;
   }
   
   timedelay = time(NULL) - timeout + 2; /* calibration for portscan */

   if ((ack_diff == 2) || (!j)) { /* is our response weird? */
      stralloc(walk, "\n\t- non-standard response, reverting to SYN scan");
      stralloc(walk, "\n\t- target likely to be a Windows based machine");
      scanfunc = synscan; /* weird tagets get SYN scaned -- more stable */
   } else scanfunc = finscan; /* go for the FIN cause our target is ok */

   if(mns_opts & SYNSCAN) scanfunc = synscan; /* forced SYN scan? */

   /* if we get too few open ports using a FIN scan, our target is suspect
    * of being extra-weird, so let's SYN them. examples of this are IRIX
    * and HP-UX but there are a few others
    */

   if((j = scanfunc(sd_send, sd_recv, pschkd, seq_ack)) < 2)
      if(scanfunc == finscan) {
         stralloc(walk, "\n\t- FIN scan may have failed, trying SYN scan");
         j = synscan(sd_send, sd_recv, pschkd, seq_ack);
      }

   stralloc (walk, "\n\tStandard TCP service ports open ->");
   if(j) {

      *plist = malloc((j + 1) * sizeof(int));
      (*plist)[0] = j;

      for (i = k = 0; i < SERVICES; i++)
	{
	  if (!pschkd[i]) {
	      k++;
	      (*plist)[k] = service[i].port;
	      sprintf (buffer, "\n\t\t%3i  tcp\t%s",
		       service[i].port, service[i].name);
	      stralloc (walk, buffer);
	    }
	}
    }
  else {
     stralloc (walk, "( none )");
     *plist = NULL;
  }

  close (sd_send);
  close (sd_recv);

  return walk;
}

struct node *checkos (struct in_addr target,
                      struct node *walk,
                      int *plist,
                      struct in_addr hostaddress)
{
   int sd_send, sd_recv;
   char *buffer = NULL;
   int i, pcount;
   char ham_asc[2];

   if (plist != NULL) pcount = plist[0] + 1;
   else return walk;

   rawsockets (&sd_send, &sd_recv);
   local_ip = hostaddress;

   for (i = 1; i < min(pcount, 3); i++) {
      dst_port = htons(plist[i]);
      buffer = checkosmain(sd_send, sd_recv);
      if (buffer != NULL) {
         ham_asc[0] = buffer[0];
	 ham_asc[1] = '\0';
	 stralloc (walk, "\n\tTCP Sequence Prediction Difficulty ->");
         stralloc (walk, ham_asc);
	 switch (ham_asc[0]) {
            case '0':
               stralloc (walk, "( indeterminate )");
	       break;
            case '1':
	       stralloc (walk, "( easy )");
	       break;
	    case '2':
	       stralloc (walk, "( easy )");
               break;
	    case '3':
	       stralloc (walk, "( moderate )");
	       break;
	    case '4':
	       stralloc (walk, "( moderate )");
	       break;
	    case '5':
	       stralloc (walk, "( hard )");
	       break;
	    case '6':
	       stralloc (walk, "( infeaseable )");
	       break;
	    case '7':
	       stralloc (walk, "( impossible )");
	       break;
	    case '8':
	       stralloc (walk, "( impossible )");
	       break;
	    default:
	       stralloc (walk, "( hamming error )");
	       break;
	    }
	    stralloc (walk, "\n\tOperating System ->");
	    stralloc (walk, &buffer[1]);
	    break;
         }
         if (buffer != NULL)
	   break;
      }

   if (buffer != NULL) free (buffer);
   else stralloc (walk, "\n\tOperating System -> ( indeterminate )");

   close (sd_send);
   close (sd_recv);

   return walk;
}

char *checkosmain (int sd_send, int sd_recv) {
  FILE *fs_cfg;
  char *osname;
  struct fingerprint pak[PACKETS];
  struct tcphdr tcp;
  hdrinfo seq_ack;
  u_short start, s;
  u_long myseq[PACKETS];
  u_long seq_mask = 0, last_seq = 0;
  time_t timeout;
  char line[60];
  int n, i, ham;

  for (n = 0; n < PACKETS; n++)
    bzero (&pak[n], sizeof (struct fingerprint));

  srand (time (NULL) & 0x0000ffff);
  start = s = (u_short) getpid () + 5000 + (rand () % 5000);
  seq_ack.ack = 0;

#define sendtcppacket(a, b)\
   {seq_ack.seq = myseq[a] = rand();\
    src_port = htons(s + a);\
    sendtcp(sd_send, &seq_ack, b);\
    usleep(10);}

  sendtcppacket (0, TH_SYN);
  sendtcppacket (1, TH_SYN | TH_ACK);
  sendtcppacket (2, TH_FIN);
  sendtcppacket (3, TH_FIN | TH_ACK);
  sendtcppacket (4, TH_SYN | TH_FIN);
  sendtcppacket (5, TH_PUSH);
  sendtcppacket (6, TH_SYN | TH_XXX | TH_YYY);
  sendtcppacket (7, TH_SYN);	/* this extra syn is used for hamming distances
				 * an even number of responses are needed */
  timeout = time (NULL) + TIMEOUT;
  n = i = 0;

  do {
     if (gettcp(sd_recv, &tcp)) {
        if (dst_port == tcp.th_sport) {
           if (ntohs(tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;
              if ((n < 0) || (n >= PACKETS)) continue;
              if (pak[n].set) continue;
              if (!n && !ntohs(tcp.th_win)) return NULL;
              pak[n].seq = tcp.th_seq != 0;
              if (pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              pak[n].ack = ntohl (tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if (pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if (pak[n].ackset)
                 pak[n].ack = (myseq[n] < pak[n].ack) ?
                 (pak[n].ack - myseq[n]) :
                 (myseq[n] - pak[n].ack);
              if (pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              pak[n].win = ntohs (tcp.th_win);
              pak[n].flag = tcp.th_flags;
	      pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep (10);
  }
  while ((timeout > time (NULL)) && (n != PACKETS - 1));
  
  /* calculate hamming distance */
  if (i % 2) seq_mask ^= last_seq;
  ham = hamming (seq_mask);
  if (ham > 8) ham = 16 - ham;

  /* parse fingerprints */
  if ((fs_cfg = fopen (OSCFGFILE, "r")) != NULL)
    {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      osname = malloc (60);

      while (fgets (line, 60, fs_cfg))
	{
	  if (line[0] == '#')
	    continue;		/* COMMENT LINE */
	  if (line[0] == '\n')
	    {			/* BLANK LINE */
	      if (flag1 && (found == total))
		{
		  fclose (fs_cfg);
		  osname[0] = (char) ham + '0';
		  return osname;
		};
	    }
	  if (line[0] == '*')
	    {			/* OS FINGERPRINT HEADER */
	      if (strlen (line) > 3)
		{
		  strcpy (osname, &line[1]);
		  osname[strlen (osname) - 1] = '\0';
		}
	      flag1 = 1;
	      total = found = 0;
	      continue;
	    }

	  /* line number */
	  token = strtok (line, " ");
	  if (isdigit (token[0]))
	    n = atoi (token);
	  else
	    {
	      flag1 = total = found = 0;
	      continue;
	    }

/*-- seq --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.seq = atoi (token);
	  parse.set = isdigit (token[0]) ? 1 : 0;

/*-- ack --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    {
	      switch (token[0])
		{
		case '+':
		  parse.ackset = 1;
		  parse.ack = atol (&token[1]);
		  break;
		case '-':
		  parse.ackset = 2;
		  parse.ack = atol (&token[1]);
		  break;
		case 'R':
		  parse.ackset = 3;
		  parse.ack = 0;
		  break;
		case 'N':
		default:
		  parse.ackset = 0;
		  parse.ack = 0;
		  break;
		}
	    }

/*-- win --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.win = strtol (token, NULL, 16);

/*-- flags --*/
	  if ((token = strtok (NULL, " \n")) != NULL)
	    {
	      parse.flag = 0;
	      if (strchr (token, 'S'))
		parse.flag |= TH_SYN;
	      if (strchr (token, 'R'))
		parse.flag |= TH_RST;
	      if (strchr (token, 'A'))
		parse.flag |= TH_ACK;
	      if (strchr (token, 'F'))
		parse.flag |= TH_FIN;
	      if (strchr (token, 'P'))
		parse.flag |= TH_PUSH;
	      if (strchr (token, 'X'))
		parse.flag |= TH_XXX;
	      if (strchr (token, 'Y'))
		parse.flag |= TH_YYY;
	      if (strchr (token, 'U'))
		parse.urg = 1;
	      else
		parse.urg = 0;
	    }

	  total++;

	  /* test for no response */
	  if (!pak[n].set)
	    {
	      if (parse.set)
		total = found = flag1 = 0;
	      else
		found++;
	      continue;
	    }

	  if ((parse.ackset == pak[n].ackset) &&
	      (parse.ack == pak[n].ack) &&
	      (parse.seq == pak[n].seq) &&
	      (parse.win == pak[n].win) &&
	      (parse.flag == pak[n].flag) &&
	      (parse.urg == pak[n].urg))
	    {
	      found++;
	      continue;
	    }
	  else
	    {
	      total = found = flag1 = 0;
	      continue;
	    }
	}

      fseek (fs_cfg, 0L, SEEK_END);
      fclose (fs_cfg);
      if (osname != NULL) free(osname);
      return NULL;
    }

  printf("error: cannot read from " OSCFGFILE "\n");
  return NULL;
}

/* Sends a TCP packet */
void sendtcp (int sd_send, hdrinfo *routing, u_short flags) {
  struct tcphdr tcp;
  SAin sa_dest;

  tcp.th_sport = src_port;
  tcp.th_dport = dst_port;
  tcp.th_seq = htonl (routing->seq);
  tcp.th_ack = 0;
  tcp.th_x2 = 0;
  tcp.th_off = 5;
  tcp.th_flags = flags;
  tcp.th_win = htons(0x1000 | (rand () % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;
/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof (struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp, sizeof(tcp), local_ip, target_ip);
#endif /* SOLARIS_CKSUM_BUG */

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = 0;
   
   if (sendto(sd_send, &tcp, sizeof(tcp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
      perror ("sending message");
}


u_long gettcp(int sd_recv, struct tcphdr *dtcp) {
   char pkt[MIN_PKT];

   if(recv(sd_recv, pkt, MIN_PKT, 0) < 0) return 0;
   if(target_ip.s_addr != iph->ip_src.s_addr) return 0;

   memcpy(dtcp, tcph, sizeof(struct tcphdr));

   return iph->ip_dst.s_addr;
}

void rawsockets (int *sd_send, int *sd_recv) {
   int rsflags;
/*-- SEND RAW socket --*/
  if ((*sd_send = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw send socket");
      exit (EXIT_FAILURE);
    }
/*-- READ RAW socket --*/
  if ((*sd_recv = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw read socket");
      exit (EXIT_FAILURE);
    }
  if ((rsflags = fcntl (*sd_recv, F_GETFL)) == -1)
    {
      perror ("fcntl(sd_recv,F_GETFL)");
      exit (EXIT_FAILURE);
    }
  if (fcntl (*sd_recv, F_SETFL, rsflags | O_NONBLOCK) == -1)
    {
      perror ("fcntl(sd_recv,F_SETFL)");
      exit (EXIT_FAILURE);
    }
}

int synscan(int sd_send, int sd_recv, int *pschkd, hdrinfo seq_ack) {
  struct tcphdr tcp;
  int i, j, found = 0;
  time_t timeout;

  for (j = 0; j < SERVICES; j++)
    pschkd[j] = PORT_CLOSED;

   for (i = 0; i < SERVICES; i++) {
      dst_port = htons(service[i].port);

      seq_ack.seq = rand ();
      sendtcp (sd_send, &seq_ack, TH_SYN);

      if(gettcp(sd_recv, &tcp))
         for (j = 0; j < SERVICES; j++)
            if(service[j].port == ntohs(tcp.th_sport)) {
               if(ntohs(tcp.th_win)) {
                  pschkd[j] = 0;
                  found++;
		  break;
               }
            }
   }

   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if (gettcp (sd_recv, &tcp)) {
         for (j = 0; j < SERVICES; j++)
            if (service[j].port == ntohs (tcp.th_sport)) {
               if (ntohs (tcp.th_win)) {
                  pschkd[j] = 0;
                  found++;
                  break;
               }
            }
      }
   }
   return found;
}

int finscan(int sd_send, int sd_recv, int *pschkd, hdrinfo seq_ack) {
   struct tcphdr tcp;
   int i, j, found = SERVICES;
   time_t timeout;
   u_short tcpflags = 0;

   for (i = 0; i < SERVICES; i++) {
      dst_port = htons(service[i].port);
      
      seq_ack.seq = rand();
      
      switch (i % 3) { /* we actually use NUL/FIN/PSH scanning */
	case 0: tcpflags = 0x0;     break; /* NULL */
	case 1: tcpflags = TH_FIN;  break; /* FIN  */
	case 2: tcpflags = TH_PUSH; break; /* PUSH */
      }
      sendtcp (sd_send, &seq_ack, tcpflags);

      if (gettcp(sd_recv, &tcp)) {
         for(j = 0; j < SERVICES; j++)
            if(service[j].port == ntohs(tcp.th_sport)) {
               pschkd[j] = PORT_CLOSED;
               found--;
               break;
            }
      }
   }
   timeout = time (NULL) + timedelay;

   while (timeout > time (NULL)) {
      if(gettcp(sd_recv, &tcp)) {
         for (j = 0; j < SERVICES; j++)
            if(service[j].port == ntohs (tcp.th_sport)) {
               pschkd[j] = PORT_CLOSED;
               found--;
               break;
            }
      }
   }
   return found;
}

struct in_addr targetup(int sd_recv) {
   SAin sa_dest;
   struct in_addr ip;
   struct tcphdr tcp;
   struct icmphdr icmp;
   int sockfd, rsflags;
   char pkt[MIN_PKT];
   time_t timeout;

   /* we use an ICMP packet to retrieve the correct interface IP from the
    * response packet for 2 reasons, to make sure the target is even up, and
    * it's easier that using ioctl(). if a ping fails, we try a connect() to
    * see if we get a refusal
    */
    
   ip.s_addr = 0;

   sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
   if((rsflags = fcntl(sockfd, F_GETFL)) == -1) {
      perror ("fcntl(sd_recv,F_GETFL)");
      exit (EXIT_FAILURE);
   }
   
   if(fcntl(sockfd, F_SETFL, rsflags | O_NONBLOCK) == -1) {
      perror ("fcntl(sd_recv,F_SETFL)");
      exit (EXIT_FAILURE);
   }

   bzero(&icmp, sizeof(icmp));
   icmp.type = ICMP_ECHO;
   icmp.un.echo.id = getpid();
   icmp.checksum = in_cksum((u_short *)&icmp, sizeof(icmp));

   sa_dest.sin_family = AF_INET;
   sa_dest.sin_addr = target_ip;
   sa_dest.sin_port = htons((rand() % 1000) + 1000);
   
   sendto(sockfd, &icmp, sizeof(icmp), 0, (SA *)&sa_dest, sizeof(SAin));

   timeout = time(NULL) + TIMEOUT;
   while(timeout > time(NULL))
      if(recv(sockfd, pkt, MIN_PKT, 0) >= 0)
         if(target_ip.s_addr == iph->ip_src.s_addr)
            if(icmph->type == ICMP_ECHOREPLY)
               if(icmph->un.echo.id == getpid()) {
                  ip = iph->ip_dst;
                  break;
               }

   /* if the ping fails let's try a sloppier connect() way */
   if(!ip.s_addr) {
      close(sockfd);
      sockfd = socket(AF_INET, SOCK_STREAM, 0);
      if((rsflags = fcntl(sockfd, F_GETFL)) == -1) {
         perror ("fcntl(sd_recv,F_GETFL)");
         exit (EXIT_FAILURE);
      }
      if(fcntl(sockfd, F_SETFL, rsflags | O_NONBLOCK) == -1) {
         perror ("fcntl(sd_recv,F_SETFL)");
         exit (EXIT_FAILURE);
      }
      connect(sockfd, (SA *)&sa_dest, sizeof(SAin));
      
      timeout = time(NULL) + TIMEOUT;
      while(timeout > time(NULL))
         if((ip.s_addr = gettcp(sd_recv, &tcp)))
            if(sa_dest.sin_port == tcp.th_sport)
               break;
   }

   close(sockfd);
   return(ip);
}
